# -*-: coding utf-8 -*-
""" Logging utilities. """

import inspect

LOGGING_ENABLED = True


class TermincalColors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'


def log(message):
    """ Print a log message.

    :param message: the message to print.
    """
    __log(inspect.stack(), message)


def log_error(message):
    """ Print an error message.

    :param message: the message to print.
    """
    __log(inspect.stack(), message, TermincalColors.FAIL)


def log_success(message):
    """ Print a success message.

    :param message: the message to print.
    """
    __log(inspect.stack(), message, TermincalColors.OKGREEN)


def log_warning(message):
    """ Print a warning message.

    :param message: the message to print.
    """
    __log(inspect.stack(), message, TermincalColors.WARNING)


def log_blue(message):
    """ Print a blue message.

    :param message: the message to print.
    """
    __log(inspect.stack(), message, TermincalColors.OKBLUE)



def debug_log(message):
    """ Print a log message.

    :param message: the message to print.
    """
    __log(inspect.stack(), message, None, True)


def debug_log_error(message):
    """ Print an error message.

    :param message: the message to print.
    """
    __log(inspect.stack(), message, TermincalColors.FAIL, True)


def debug_log_success(message):
    """ Print a success message.

    :param message: the message to print.
    """
    __log(inspect.stack(), message, TermincalColors.OKGREEN, True)


def debug_log_warning(message):
    """ Print a warning message.

    :param message: the message to print.
    """
    __log(inspect.stack(), message, TermincalColors.WARNING, True)


def debug_log_blue(message):
    """ Print a blue message.

    :param message: the message to print.
    """
    __log(inspect.stack(), message, TermincalColors.OKBLUE, True)

def __log(stack, message, color=None, include_tag=False):
    """ Print a message, optionally in a given color.

    :param message: the message to print.
    :param color: the color of the message.
    """
    if not LOGGING_ENABLED or message is None:
        return

    frame = stack[1][0]
    caller = frame.f_locals.get('self', None)
    class_name = frame.f_locals.get('cls', None)
    
    if caller is not None:
        tag = caller.__class__.__name__
    elif class_name is not None:
        tag = class_name.__name__
    else:
        tag = "LOG"

    if include_tag and tag is not None and len(tag) > 0:
        message = "[{}] {}".format(tag, message)

    if not color:
        print(message)
    else:
        print("{}{}{}".format(color, message, TermincalColors.ENDC))



# -*-: coding utf-8 -*-
""" Helper class for parsing intents. """

from dateutil.parser import parse

from snipsskillscore.instant_time import InstantTime
from snipsskillscore.time_interval import TimeInterval


class IntentParser:
    """ Helper class for parsing intents. """

    @staticmethod
    def parse(payload, candidate_classes):
        """ Parse a json response into an intent.

        :param payload: a JSON object representing an intent.
        :param candidate_classes: a list of classes representing various
                                  intents, each having their own `parse`
                                  method to attempt parsing the JSON object
                                  into the given intent class.
        :return: An object version of the intent if one of the candidate
                 classes managed to parse it, or None.
        """
        for cls in candidate_classes:
            intent = cls.parse(payload)
            if intent:
                return intent
        return None

    @staticmethod
    def get_intent_name(payload):
        """ Return the simple intent name. An intent has the form:

           {
               "input": "turn the lights green",
               "intent": {
                   "intentName": "user_BJW0GIoCx__Lights",
                   ...
               },
               "slots": [...]
           }

           and this function extracts the last part of the intent
           name ("Lights"), i.e. removing the user id.

        :param payload: the intent, in JSON format.
        :return: the simpe intent name.
        """
        if 'intent' in payload and 'intentName' in payload['intent']:
            return payload['intent']['intentName'].split('__')[-1]
        return None

    @staticmethod
    def get_slot_value(payload, slot_name):
        """ Return the parsed value of a slot. An intent has the form:

            {
                "text": "brew me a cappuccino with 3 sugars tomorrow",
                "slots": [
                    {"value": {"slotName": "coffee_type", "value": "cappuccino"}},
                    ...
                ]
            }

            This function extracts a slot value given its slot name, and parses
            it into a Python object if applicable (e.g. for dates).

            Slots can be of various forms, the simplest being just:

            {"slotName": "coffee_sugar_amout", "value": "3"}

            More complex examples are date times, where we distinguish between
            instant times, or intervals. Thus, a slot:

            {
              "slotName": "weatherForecastStartDatetime",
              "value": {
                "kind": "InstantTime",
                "value": {
                  "value": "2017-07-14 00:00:00 +00:00",
                  "grain": "Day",
                  "precision": "Exact"
                }
              }
            }

            will be extracted as an `InstantTime` object, with datetime parsed
            and granularity set.

            Another example is a time interval:

            {
              "slotName": "weatherForecastStartDatetime",
              "value": {
                "kind": "TimeInterval",
                "value": {
                  "from": "2017-07-14 12:00:00 +00:00",
                  "to": "2017-07-14 19:00:00 +00:00"
                }
              },
            }

            which will be extracted as a TimeInterval object.

        :param payload: the intent, in JSON format.
        :return: the parsed value, as described above.
        """

        if not 'slots' in payload:
            return None

        slot = None
        for candidate in payload['slots']:
            if 'slotName' in candidate and candidate['slotName'] == slot_name:
                slot = candidate
                break

        if not slot:
            return None

        kind = IntentParser.get_dict_value(slot, ['value', 'kind'])
        if kind == "InstantTime":
            return IntentParser.parse_instant_time(slot)
        elif kind == "TimeInterval":
            return IntentParser.parse_time_interval(slot)

        return IntentParser.get_dict_value(slot, ['value', 'value', 'value']) \
            or IntentParser.get_dict_value(slot, ['value', 'value'])

    @staticmethod
    def parse_instant_time(slot):
        """ Parse a slot into an InstantTime object.

        :param slot: a intent slot.
        :return: a parsed InstantTime object, or None.
        """
        date = IntentParser.get_dict_value(slot, ['value', 'value', 'value'])
        if not date:
            return None
        date = parse(date)
        if not date:
            return None
        grain = InstantTime.parse_grain(
            IntentParser.get_dict_value(slot,
                                        ['value', 'value', 'grain']))
        return InstantTime(date, grain)

    @staticmethod
    def parse_time_interval(slot):
        """ Parse a slot into a TimeInterval object.

        :param slot: a intent slot.
        :return: a parsed TimeInterval object, or None.
        """
        start = IntentParser.get_dict_value(
            slot, ['value', 'value', 'from'])
        end = IntentParser.get_dict_value(slot, ['value', 'value', 'to'])
        if not start or not end:
            return None
        start = parse(start)
        end = parse(end)
        if not start or not end:
            return None
        return TimeInterval(start, end)

    @staticmethod
    def get_dict_value(dictionary, path):
        """ Safely get the value of a dictionary given a key path. For
            instance, for the dictionary `{ 'a': { 'b': 1 } }`, the value at
            key path ['a'] is { 'b': 1 }, at key path ['a', 'b'] is 1, at
            key path ['a', 'b', 'c'] is None.

        :param dictionary: a dictionary.
        :param path: the key path.
        :return: The value of d at the given key path, or None if the key
                 path does not exist.
        """
        if len(path) == 0:
            return None
        temp_dictionary = dictionary
        try:
            for k in path:
                temp_dictionary = temp_dictionary[k]
            return temp_dictionary
        except (KeyError, TypeError):
            pass
        return None



# -*-: coding utf-8 -*-
""" A simple audio player based on pygame. """

import pygame
import threading

from .logging import debug_log_warning


class AudioPlayer:
    """ A simple audio player based on pygame. """

    @classmethod
    def play(cls, file_path, on_done=None):
        """ Play an audio file.

        :param file_path: the path to the file to play.
        :param on_done: callback when audio playback completes.
        """
        pygame.mixer.init()
        try:
            pygame.mixer.music.load(file_path)
        except pygame.error as e:
            debug_log_warning(str(e))
            return

        pygame.mixer.music.play()
        while pygame.mixer.music.get_busy():
            continue
        if on_done:
            on_done()

    @classmethod
    def play_async(cls, file_path, on_done=None):
        """ Play an audio file asynchronously.

        :param file_path: the path to the file to play.
        :param on_done: callback when audio playback completes.
        """
        thread = threading.Thread(
            target=AudioPlayer.play, args=(file_path, on_done,))
        thread.start()

    @classmethod
    def stop(cls):
        """ Stop the audio. """
        pygame.mixer.init()
        pygame.mixer.music.stop()

    @classmethod
    def pause(cls):
        """ Pause the audio. """
        pygame.mixer.init()
        pygame.mixer.music.pause()

    @classmethod
    def resume(cls):
        """ Resume the audio. """
        pygame.mixer.init()
        pygame.mixer.music.unpause()



# -*-: coding utf-8 -*-
""" USB utilities. """

from __future__ import absolute_import
import os
import re
import subprocess
import logging

import usb.core
import usb.util


class USB:

    class Device:
        unknown, respeaker, conexant = range(3)

    # this method should work in all platorms
    @staticmethod
    def get_boards():
        print("+++++++++++++++ 1")
        all_devices = usb.core.find(find_all=True)

        print("+++++++++++++++ 2")
        if not all_devices:
            return USB.Device.unknown

        for board in all_devices:
            print(str(board))
            try:
                devices = board.product.lower()
                if devices.find("respeaker") >= 0:
                    return USB.Device.respeaker
                elif devices.find("conexant") >= 0:
                    return USB.Device.conexant
                else:
                    return USB.Device.unknown
            except Exception as e:
                logging.warning("Exception getting product string: %s", e)
                continue

        return USB.Device.unknown

    @staticmethod
    def get_usb_led_device():
        devices = USB.lsusb()
        if not devices:
            return None

        devices = devices.lower()

        if devices.find("respeaker") >= 0:
            return USB.Device.respeaker
        elif devices.find("conexant") >= 0:
            # TODO: check if this is the string to look for
            return USB.Device.conexant

        return USB.Device.unknown

    @staticmethod
    def lsusb():
        FNULL = open(os.devnull, 'w')
        try:
            # Raspberry
            return subprocess.check_output(["lsusb"])
        except:
            try:
                # OSX
                return subprocess.check_output(["system_profiler", "SPUSBDataType"])
            except:
                return None



# -*-: coding utf-8 -*-
""" Wrapper for various TTS services (currently only Google). """

import json
import os
import time

import paho.mqtt.client as mqtt

from socket import error as socket_error
from gtts import gTTS

from .logging import debug_log

from .audio_player import AudioPlayer


class SnipsTTS:
    """ Snips TTS service. """

    def __init__(self, thread_handler, mqtt_hostname, mqtt_port, mqtt_topic, locale):
        """ Initialise the service.

        :param mqtt_hostname: the MQTT broker hostname.
        :param mqtt_port: the MQTT broker port.
        :param mqtt_topic: the topic on which to post the sentence.
        :param locale: the language locale, e.g. "fr" or "en_US".
        """
        self.locale = locale.split("_")[0]

        self.mqtt_topic = mqtt_topic
        self.mqtt_client = mqtt.Client("SnipsTTS")
        
        thread_handler.run(target=self.start_blocking, args=(mqtt_hostname, mqtt_port, ))

    def start_blocking(self, mqtt_hostname, mqtt_port, run_event):
        while True and run_event.is_set():
            try:
                self.mqtt_client.connect(mqtt_hostname, mqtt_port)
                break
            except (socket_error, Exception) as e:
                time.sleep(5)

    def speak(self, sentence):
        """ Speak a sentence using Snips TTS.

        :param sentence: the sentence to speak.
        """
        if self.mqtt_client is None:
            return

        self.mqtt_client.publish(
            self.mqtt_topic,
            payload=json.dumps({'text': sentence}),
            qos=0,
            retain=False)


class GTTS:
    """ Google TTS service. """

    def __init__(self, locale):
        """ Initialise the service.

        :param locale: the language locale, e.g. "fr" or "en_US".
        """
        self.locale = locale.split("_")[0]

    def speak(self, sentence):
        """ Speak a sentence using Google TTS.

        :param sentence: the sentence to speak.
        """
        snips_dir = ".snips"
        filename = "gtts.mp3"
        file_path = "{}/{}".format(snips_dir, filename)

        if not os.path.exists(snips_dir):
            os.makedirs(snips_dir)

        def delete_file():
            try:
                os.remove(file_path)
                if not os.listdir(snips_dir):
                    try:
                        os.rmdir(snips_dir)
                    except OSError:
                        pass
            except:
                pass

        debug_log("Speak: {}".format(sentence))
        tts = gTTS(text=sentence, lang=self.locale)
        tts.save(file_path)
        AudioPlayer.play_async(file_path, delete_file)



from setuptools import setup

setup(
    name='snipsskillscore',
    version='0.1.5.5.40',
    description='The Snips skills core utilities for creating end-to-end assistants',
    author='Michael Fester',
    author_email='michael.fester@gmail.com',
    url='https://github.com/snipsco/snips-skills-core',
    download_url='',
    license='MIT',
    install_requires=[
        'webrtcvad',
        'cython',
        'hidapi',
        'paho-mqtt',
        'pyyaml',
        'pyaudio',
        'pygame',
        'pyusb',
        'snips-respeaker',
        'gTTS'
    ],
    test_suite="tests",
    keywords=['snips'],
    packages=[
        'snipsskillscore'
    ],
    include_package_data=True,
)



# -*-: coding utf-8 -*-
""" Handler for various states of the system. """

from .sound_service import SoundService
from .leds_service import LedsService

class State:
    none, welcome, goodbye, hotword_toggle_on, hotword_detected, asr_toggle_on, asr_text_captured, error, idle = range(
        9)

class StateHandler:
    """ Handler for various states of the system. """

    def __init__(self, thread_handler):
        self.leds_service = LedsService(thread_handler)
        self.state = None

    def set_state(self, state):
        if state == State.welcome:
            SoundService.play(SoundService.State.welcome)
        elif state == State.goodbye:
            SoundService.play(SoundService.State.goodbye)
            self.leds_service.start_animation(LedsService.State.none)
        elif state == State.hotword_toggle_on and self.state != state:
            self.leds_service.start_animation(LedsService.State.standby)
        elif state == State.hotword_detected:
            SoundService.play(SoundService.State.hotword_detected)
        elif state == State.asr_toggle_on and self.state != state:
            self.leds_service.start_animation(LedsService.State.listening)
        elif state == State.asr_text_captured:
            SoundService.play(SoundService.State.asr_text_captured)
        elif state == State.error:
            self.leds_service.start_animation(LedsService.State.error)
            SoundService.play(SoundService.State.error)
        self.state = state



# -*-: coding utf-8 -*-
""" Leds service for handling visual feedback for various states of
    the system. """

import random
import time
import struct
import threading

from usb.core import USBError

from .singleton import Singleton
from .usb_utils import USB
from .logging import log_error

try:
    import respeaker.usb_hid as usb_hid
except USBError:
    log_error("Error accessing microphone: insufficient permissions. " +
              "You may need to replug your microphone and restart your device.")


class LedsService:
    """ Leds service for handling visual feedback for various states of
        the system. """

    class State:
        """ Leds states.- """
        none, waking_up, standby, listening, loading, notify, error = range(7)

    def __init__(self, thread_handler):
        self.thread_handler = thread_handler
        led_device = USB.get_boards()
        if led_device == USB.Device.respeaker:
            self.animator = ReSpeakerAnimator()
        else:
            self.animator = None

    def start_animation(self, animation_id):
        if not self.animator:
            return
        (animation, identifier) = self.get_animation(animation_id)

        self.thread_handler.run(target=self.animator.run,
                                args=(identifier, animation, ))

    def get_animation(self, animation):
        identifier = str(random.randint(1, 100000))
        return (Animation(identifier, animation), identifier)


class Animation(Singleton):

    def __init__(self, id, active=0):
        self.id = id
        self.active = active


class ReSpeakerAnimator(object):

    def __init__(self):
        self.hid = usb_hid.get()
        self.led_dict = {
            'LED_1': 3,
            'LED_2': 4,
            'LED_3': 4,
            'LED_4': 5,
            'LED_5': 6,
            'LED_6': 7,
            'LED_7': 8,
            'LED_8': 9,
            'LED_9': 10,
            'LED_10': 11,
            'LED_11': 12,
            'LED_12': 13,
            'LED_13': 14,
        }

        self.animation_none = [
            {value: 0 for value in self.led_dict.values()}]
        self.animation_waking_up = [
            {value: 2 * st for value in self.led_dict.values()} for st in range(127)]
        self.animation_standby = [{value: st for value in self.led_dict.values(
        )} for st in list(range(31)) + list(reversed(range(31)))]

        # the matrix above:
        def value(i, j, N):
            if i == j:
                return 0
            elif i == (j + 1) % N:
                return 0
            elif i == (j - 1) % N:
                return 64
            elif i == (j + 2) % N or i == (j - 2) % N:
                return 127
            else:
                return 255

        self.animation_listening = []
        leds = range(3, 15)
        N = len(leds)

        for i in range(0, N):
            self.animation_listening.append(
                {leds[j]: value(i, j, N) for j in range(0, N)})

        self.animation_loading = [{value: 16 * st for value in self.led_dict.values()}
                                  for st in list(range(15)) + list(reversed(range(15)))]

        self.animation_notify = []
        for frame in range(0, 2):
            self.animation_notify.append(
                {value: int("0000ff", 16) for value in self.led_dict.values()})
            self.animation_notify.append(
                {value: int("000000", 16) for value in self.led_dict.values()})

        self.animation_error = []
        for frame in range(0, 3):
            self.animation_error.append(
                {value: int("ff0000", 16) for value in self.led_dict.values()})
            self.animation_error.append(
                {value: int("000000", 16) for value in self.led_dict.values()})

        # Set the ReSpeaker in the right mode
        try:
            init_addr = 0  # 0x0
            init_data = 6  # 0x00000006

            self.write(init_addr, init_data)
        except Exception as e:
            print(e.message)

    def write(self, address, data):
        if data > 0xFFFF:
            data = struct.pack('<I', data)
        elif data > 0xFF:
            data = struct.pack('<H', data)

        data = self.to_bytearray(data)
        length = len(data)
        if self.hid:
            packet = bytearray([address & 0xFF, (address >> 8) &
                                0x7F, length & 0xFF, (length >> 8) & 0xFF]) + data
            packet = list(packet)
            self.hid.write(packet)

    def read(self, address, length):
        if self.hid:
            self.hid.write(list(bytearray(
                [address & 0xFF, (address >> 8) & 0xFF | 0x80, length & 0xFF, (length >> 8) & 0xFF])))
            for _ in range(6):
                data = self.hid.read()
                # skip VAD data
                if int(data[0]) != 0xFF and int(data[1]) != 0xFF:
                    return data[4:(4 + length)]

    @staticmethod
    def to_bytearray(data):
        if type(data) is int:
            array = bytearray([data & 0xFF])
        elif type(data) is bytearray:
            array = data
        elif type(data) is str:
            array = bytearray(data)
        elif type(data) is list:
            array = bytearray(data)
        else:
            raise TypeError('%s is not supported' % type(data))

        return array

    def run(self, id, animation, run_event):
        if animation.active == LedsService.State.none:
            for item in self.animation_none:
                if animation.id != id or not run_event.is_set():
                    break
                for k, v in item.items():
                    self.write(k, v)
                time.sleep(0.5)
        elif animation.active == LedsService.State.waking_up:
            for item in self.animation_waking_up:
                if animation.id != id or not run_event.is_set():
                    break
                for k, v in item.items():
                    self.write(k, v)
                time.sleep(0.02)

        elif animation.active == LedsService.State.standby:
            while True:
                if animation.id != id or not run_event.is_set():
                    break
                for item in self.animation_standby:
                    if animation.id != id or not run_event.is_set():
                        break
                    for k, v in item.items():
                        self.write(k, v)
                    time.sleep(0.05)
        elif animation.active == LedsService.State.listening:
            while True:
                if animation.id != id or not run_event.is_set():
                    break
                for item in self.animation_listening:
                    if animation.id != id or not run_event.is_set():
                        break
                    for k, v in item.items():
                        self.write(k, v)
                    time.sleep(0.01)
        elif animation.active == LedsService.State.loading:
            while True:
                if animation.id != id or not run_event.is_set():
                    break
                for item in self.animation_loading:
                    if animation.id != id or not run_event.is_set():
                        break
                    for k, v in item.items():
                        self.write(k, v)
                    time.sleep(0.01)
        elif animation.active == LedsService.State.notify:
            for item in self.animation_notify:
                if animation.id != id or not run_event.is_set():
                    break
                for k, v in item.items():
                    self.write(k, v)
                time.sleep(0.1)
        elif animation.active == LedsService.State.error:
            for item in self.animation_error:
                if animation.id != id or not run_event.is_set():
                    break
                for k, v in item.items():
                    self.write(k, v)
                time.sleep(0.2)
